package com.virjar.nat.client;

import com.virjar.nat.client.listener.ChannelStatusListener;
import com.virjar.nat.handlers.ClientChannelHandler;
import com.virjar.nat.handlers.LittleProxyServerChannelHandler;
import com.virjar.nat.protocol.ClientIdleCheckHandler;
import com.virjar.nat.protocol.ProxyMessage;
import com.virjar.nat.protocol.ProxyMessageDecoder;
import com.virjar.nat.protocol.ProxyMessageEncoder;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLEngine;

import io.netty.bootstrap.Bootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.ssl.SslHandler;


public class ProxyClient implements ChannelStatusListener {

    private static Logger logger = LoggerFactory.getLogger(ProxyClient.class);

    private static final int MAX_FRAME_LENGTH = 1024 * 1024;

    private static final int LENGTH_FIELD_OFFSET = 0;

    private static final int LENGTH_FIELD_LENGTH = 4;

    private static final int INITIAL_BYTES_TO_STRIP = 0;

    private static final int LENGTH_ADJUSTMENT = 0;

    private NioEventLoopGroup workerGroup;

    private Bootstrap join2NatServerBootStrap;

    private Bootstrap join2LittleProxyBootStrap;


    private static long sleepTimeMill = 1000;

    private String serverHost;
    private int serverPort;
    private String clientID;

    private ClientChannelMannager clientChannelMannager;

    public ProxyClient(String serverHost, int serverPort, String clientID) {
        this.serverHost = serverHost;
        this.serverPort = serverPort;
        this.clientID = clientID;

        clientChannelMannager = new ClientChannelMannager(serverHost);
    }

    public ClientChannelMannager getClientChannelMannager() {
        return clientChannelMannager;
    }

    /**
     * 启动内网穿透客户端
     *
     * @param serverHost 公网服务主机
     * @param serverPort 公网服务端口
     * @param clientID   客户端标记，同一个客户端标记在服务器端会映射为同一个端口
     */
    public static ProxyClient start(String serverHost, int serverPort, final String clientID) {
        ProxyClient proxyClient = new ProxyClient(serverHost, serverPort, clientID);
        proxyClient.startInernal();
        return proxyClient;
    }


    public void startInernal() {
        workerGroup = new NioEventLoopGroup();
        join2LittleProxyBootStrap = new Bootstrap();
        join2LittleProxyBootStrap.group(workerGroup);
        join2LittleProxyBootStrap.channel(NioSocketChannel.class);
        join2LittleProxyBootStrap.handler(new ChannelInitializer<SocketChannel>() {

            @Override
            public void initChannel(SocketChannel ch) throws Exception {
                ch.pipeline().addLast(new LittleProxyServerChannelHandler(ProxyClient.this));
            }
        });

        join2NatServerBootStrap = new Bootstrap();
        join2NatServerBootStrap.group(workerGroup);
        join2NatServerBootStrap.channel(NioSocketChannel.class);
        join2NatServerBootStrap.handler(new ChannelInitializer<SocketChannel>() {

            @Override
            public void initChannel(SocketChannel ch) throws Exception {
                ch.pipeline().addLast(new ProxyMessageDecoder(MAX_FRAME_LENGTH, LENGTH_FIELD_OFFSET, LENGTH_FIELD_LENGTH, LENGTH_ADJUSTMENT, INITIAL_BYTES_TO_STRIP));
                ch.pipeline().addLast(new ProxyMessageEncoder());
                ch.pipeline().addLast(new ClientIdleCheckHandler());
                ch.pipeline().addLast(new ClientChannelHandler(join2LittleProxyBootStrap, join2NatServerBootStrap, ProxyClient.this, ProxyClient.this, clientID));
            }
        });
        connectProxyServer();


    }


    private ChannelHandler createSslHandler(SSLContext sslContext) {
        SSLEngine sslEngine = sslContext.createSSLEngine();
        sslEngine.setUseClientMode(true);
        return new SslHandler(sslEngine);
    }

    private void connectProxyServer() {
        join2NatServerBootStrap.connect(serverHost, serverPort).addListener(new ChannelFutureListener() {
            @Override
            public void operationComplete(ChannelFuture future) throws Exception {
                if (future.isSuccess()) {

                    // 连接成功，向服务器发送客户端认证信息（clientKey）
                    clientChannelMannager.setCmdChannel(future.channel());
                    ProxyMessage proxyMessage = new ProxyMessage();
                    proxyMessage.setType(ProxyMessage.C_TYPE_AUTH);
                    proxyMessage.setUri(clientID);
                    future.channel().writeAndFlush(proxyMessage);
                    sleepTimeMill = 1000;
                    logger.info("connect proxy server success, {}", future.channel());
                } else {
                    logger.warn("connect proxy server failed", future.cause());
                    // 连接失败，发起重连
                    reconnectWait();
                    connectProxyServer();
                }
            }
        });
    }

    public void stop() {
        workerGroup.shutdownGracefully();
    }

    public void channelInactive(ChannelHandlerContext ctx) {
        reconnectWait();
        connectProxyServer();
    }

    private static void reconnectWait() {
        try {
            if (sleepTimeMill > 60000) {
                sleepTimeMill = 1000;
            }

            synchronized (ProxyClient.class) {
                sleepTimeMill = sleepTimeMill * 2;
                Thread.sleep(sleepTimeMill);
            }
        } catch (InterruptedException e) {
        }
    }

}
